use crate::error::Result;
use serde::{Deserialize, Serialize};
use std::{fs::File, io::Read, path::PathBuf};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct DioxusConfig {
    pub application: ApplicationConfig,
    pub web: WebConfig,
}

impl DioxusConfig {
    pub fn load() -> crate::error::Result<DioxusConfig> {
        let crate_dir = crate::cargo::crate_root()?;

        if !crate_dir.join("Dioxus.toml").is_file() {
            return Err(crate::error::Error::Unique("Config file: `Dioxus.toml` not found.".into()));
        }

        let mut dioxus_conf_file = File::open(crate_dir.join("Dioxus.toml"))?;
        let mut meta_str = String::new();
        dioxus_conf_file.read_to_string(&mut meta_str)?;

        match toml::from_str::<DioxusConfig>(&meta_str) {
            Ok(v) => return Ok(v),
            Err(e) => {
                log::error!("{}", e);
                return Err(crate::error::Error::Unique(
                    "Dioxus.toml parse failed".into(),
                ));
            }
        }
    }
}

impl Default for DioxusConfig {
    fn default() -> Self {
        Self {
            application: ApplicationConfig {
                name: "dioxus".into(),
                platforms: vec![String::from("web")],
                out_dir: Some(PathBuf::from("dist")),
            },
            web: WebConfig {
                app: WebAppConfing {
                    title: Some("dioxus | ⛺".into()),
                    public_dir: Some(PathBuf::from("public")),
                },
                watcher: WebWatcherConfing {
                    watch_path: Some(vec![PathBuf::from("src")]),
                    reload_html: Some(false),
                },
                resource: WebResourceConfing {
                    dev: WebDevResourceConfing {
                        style: Some(vec![]),
                        script: Some(vec![]),
                    },
                    style: Some(vec![]),
                    script: Some(vec![]),
                },
            },
        }
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ApplicationConfig {
    pub name: String,
    pub platforms: Vec<String>,
    pub out_dir: Option<PathBuf>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct WebConfig {
    pub app: WebAppConfing,
    pub watcher: WebWatcherConfing,
    pub resource: WebResourceConfing,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct WebAppConfing {
    pub title: Option<String>,
    pub public_dir: Option<PathBuf>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct WebWatcherConfing {
    pub watch_path: Option<Vec<PathBuf>>,
    pub reload_html: Option<bool>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct WebResourceConfing {
    pub dev: WebDevResourceConfing,
    pub style: Option<Vec<PathBuf>>,
    pub script: Option<Vec<PathBuf>>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct WebDevResourceConfing {
    pub style: Option<Vec<PathBuf>>,
    pub script: Option<Vec<PathBuf>>,
}

#[derive(Debug, Clone)]
pub struct CrateConfig {
    pub out_dir: PathBuf,
    pub crate_dir: PathBuf,
    pub workspace_dir: PathBuf,
    pub target_dir: PathBuf,
    pub public_dir: PathBuf,
    pub manifest: cargo_toml::Manifest<cargo_toml::Value>,
    pub executable: ExecutableType,
    pub dioxus_config: DioxusConfig,
    pub release: bool,
}

#[derive(Debug, Clone)]
pub enum ExecutableType {
    Binary(String),
    Lib(String),
    Example(String),
}

impl CrateConfig {
    pub fn new() -> Result<Self> {
        let dioxus_config = DioxusConfig::load()?;

        let crate_dir = crate::cargo::crate_root()?;
        let workspace_dir = crate::cargo::workspace_root()?;
        let target_dir = workspace_dir.join("target");

        let out_dir = match dioxus_config.application.out_dir {
            Some(ref v) => crate_dir.join(v),
            None => crate_dir.join("dist"),
        };

        let cargo_def = &crate_dir.join("Cargo.toml");

        let public_dir = match dioxus_config.web.app.public_dir {
            Some(ref v) => crate_dir.join(v),
            None => crate_dir.join("public"),
        };

        let manifest = cargo_toml::Manifest::from_path(&cargo_def).unwrap();

        // We just assume they're using a 'main.rs'
        // Anyway, we've already parsed the manifest, so it should be easy to change the type
        let output_filename = (&manifest)
            .lib
            .as_ref()
            .and_then(|lib| lib.name.clone())
            .or_else(|| {
                (&manifest)
                    .package
                    .as_ref()
                    .and_then(|pkg| Some(pkg.name.clone()))
                    .clone()
            })
            .expect("No lib found from cargo metadata");
        let executable = ExecutableType::Binary(output_filename);

        let release = false;

        Ok(Self {
            out_dir,
            crate_dir,
            workspace_dir,
            target_dir,
            public_dir,
            manifest,
            executable,
            release,
            dioxus_config,
        })
    }

    pub fn as_example(&mut self, example_name: String) -> &mut Self {
        self.executable = ExecutableType::Example(example_name);
        self
    }

    pub fn with_release(&mut self, release: bool) -> &mut Self {
        self.release = release;
        self
    }

    // pub fn with_build_options(&mut self, options: &BuildOptions) {
    //     if let Some(name) = &options.example {
    //         self.as_example(name.clone());
    //     }
    //     self.release = options.release;
    //     self.out_dir = options.outdir.clone().into();
    // }

    // pub fn with_develop_options(&mut self, options: &DevelopOptions) {
    //     if let Some(name) = &options.example {
    //         self.as_example(name.clone());
    //     }
    //     self.release = options.release;
    //     self.out_dir = tempfile::Builder::new().tempdir().expect("").into_path();
    // }
}
